@@ -46,40 +46,22 @@ try:
 except ImportError:
     from typing_extensions import NoReturn
 
-HAVE_ANNOTATED = sys.version_info >= (3, 7, 0)
-
 # This is a messy, type-wise, because we not only have two potentially conflicting imports here
-# but we also define our own versions in the else block in case we operate on Python 3.6
-# which didn't get Annotated support in get_type_hints(). The easiest way to make mypy
-# happy here is to tell it the versions from typing_extensions are canonical. Since this
-# typing_extensions import is only for mypy it'll work even without typing_extensions actually
-# installed so all's good.
+# The easiest way to make mypy happy here is to tell it the versions from typing_extensions are
+# canonical. Since this typing_extensions import is only for mypy it'll work even without
+# typing_extensions actually installed so all's good.
 if TYPE_CHECKING:
     from typing_extensions import _AnnotatedAlias, Annotated, get_type_hints
-elif HAVE_ANNOTATED:
+else:
     # Ignoring errors here as typing_extensions stub doesn't know about those things yet
     try:
         from typing import _AnnotatedAlias, Annotated, get_type_hints
     except ImportError:
         from typing_extensions import _AnnotatedAlias, Annotated, get_type_hints
-else:
-
-    class Annotated:
-        pass
-
-    from typing import get_type_hints as _get_type_hints
-
-    def get_type_hints(
-        obj: Callable[..., Any],
-        globalns: Optional[Dict[str, Any]] = None,
-        localns: Optional[Dict[str, Any]] = None,
-        include_extras: bool = False,
-    ) -> Dict[str, Any]:
-        return _get_type_hints(obj, globalns, localns)
 
 
 __author__ = 'Alec Thomas <alec@swapoff.org>'
-__version__ = '0.19.0'
+__version__ = '0.20.1'
 __version_tag__ = ''
 
 log = logging.getLogger('injector')
@@ -119,86 +101,85 @@ lock = threading.RLock()
 _inject_marker = object()
 _noinject_marker = object()
 
-if HAVE_ANNOTATED:
-    InjectT = TypeVar('InjectT')
-    Inject = Annotated[InjectT, _inject_marker]
-    """An experimental way to declare injectable dependencies utilizing a `PEP 593`_ implementation
-    in Python 3.9 and backported to Python 3.7+ in `typing_extensions`.
+InjectT = TypeVar('InjectT')
+Inject = Annotated[InjectT, _inject_marker]
+"""An experimental way to declare injectable dependencies utilizing a `PEP 593`_ implementation
+in Python 3.9 and backported to Python 3.7+ in `typing_extensions`.
 
-    Those two declarations are equivalent::
+Those two declarations are equivalent::
 
-        @inject
-        def fun(t: SomeType) -> None:
-            pass
+    @inject
+    def fun(t: SomeType) -> None:
+        pass
 
-        def fun(t: Inject[SomeType]) -> None:
-            pass
+    def fun(t: Inject[SomeType]) -> None:
+        pass
 
-    The advantage over using :func:`inject` is that if you have some noninjectable parameters
-    it may be easier to spot what are they. Those two are equivalent::
+The advantage over using :func:`inject` is that if you have some noninjectable parameters
+it may be easier to spot what are they. Those two are equivalent::
 
-        @inject
-        @noninjectable('s')
-        def fun(t: SomeType, s: SomeOtherType) -> None:
-            pass
+    @inject
+    @noninjectable('s')
+    def fun(t: SomeType, s: SomeOtherType) -> None:
+        pass
 
-        def fun(t: Inject[SomeType], s: SomeOtherType) -> None:
-            pass
+    def fun(t: Inject[SomeType], s: SomeOtherType) -> None:
+        pass
 
-    .. seealso::
+.. seealso::
 
-        Function :func:`get_bindings`
-            A way to inspect how various injection declarations interact with each other.
+    Function :func:`get_bindings`
+        A way to inspect how various injection declarations interact with each other.
 
-    .. versionadded:: 0.18.0
-    .. note:: Requires Python 3.7+.
-    .. note::
+.. versionadded:: 0.18.0
+.. note:: Requires Python 3.7+.
+.. note::
 
-        If you're using mypy you need the version 0.750 or newer to fully type-check code using this
-        construct.
+    If you're using mypy you need the version 0.750 or newer to fully type-check code using this
+    construct.
 
-    .. _PEP 593: https://www.python.org/dev/peps/pep-0593/
-    .. _typing_extensions: https://pypi.org/project/typing-extensions/
-    """
+.. _PEP 593: https://www.python.org/dev/peps/pep-0593/
+.. _typing_extensions: https://pypi.org/project/typing-extensions/
+"""
 
-    NoInject = Annotated[InjectT, _noinject_marker]
-    """An experimental way to declare noninjectable dependencies utilizing a `PEP 593`_ implementation
-    in Python 3.9 and backported to Python 3.7+ in `typing_extensions`.
+NoInject = Annotated[InjectT, _noinject_marker]
+"""An experimental way to declare noninjectable dependencies utilizing a `PEP 593`_ implementation
+in Python 3.9 and backported to Python 3.7+ in `typing_extensions`.
 
-    Since :func:`inject` declares all function's parameters to be injectable there needs to be a way
-    to opt out of it. This has been provided by :func:`noninjectable` but `noninjectable` suffers from
-    two issues:
+Since :func:`inject` declares all function's parameters to be injectable there needs to be a way
+to opt out of it. This has been provided by :func:`noninjectable` but `noninjectable` suffers from
+two issues:
 
-    * You need to repeat the parameter name
-    * The declaration may be relatively distance in space from the actual parameter declaration, thus
-      hindering readability
+* You need to repeat the parameter name
+* The declaration may be relatively distance in space from the actual parameter declaration, thus
+  hindering readability
 
-    `NoInject` solves both of those concerns, for example (those two declarations are equivalent)::
+`NoInject` solves both of those concerns, for example (those two declarations are equivalent)::
 
-        @inject
-        @noninjectable('b')
-        def fun(a: TypeA, b: TypeB) -> None:
-            pass
+    @inject
+    @noninjectable('b')
+    def fun(a: TypeA, b: TypeB) -> None:
+        pass
 
-        @inject
-        def fun(a: TypeA, b: NoInject[TypeB]) -> None:
-            pass
+    @inject
+    def fun(a: TypeA, b: NoInject[TypeB]) -> None:
+        pass
 
-    .. seealso::
+.. seealso::
 
-        Function :func:`get_bindings`
-            A way to inspect how various injection declarations interact with each other.
+    Function :func:`get_bindings`
+        A way to inspect how various injection declarations interact with each other.
 
-    .. versionadded:: 0.18.0
-    .. note:: Requires Python 3.7+.
-    .. note::
+.. versionadded:: 0.18.0
+.. note:: Requires Python 3.7+.
+.. note::
 
-        If you're using mypy you need the version 0.750 or newer to fully type-check code using this
-        construct.
+    If you're using mypy you need the version 0.750 or newer to fully type-check code using this
+    construct.
 
-    .. _PEP 593: https://www.python.org/dev/peps/pep-0593/
-    .. _typing_extensions: https://pypi.org/project/typing-extensions/
-    """
+.. _PEP 593: https://www.python.org/dev/peps/pep-0593/
+.. _typing_extensions: https://pypi.org/project/typing-extensions/
+"""
 
 
 def reraise(original: Exception, exception: Exception, maximum_frames: int = 1) -> NoReturn:
@@ -651,8 +632,9 @@ class Binder:
 
     def get_binding(self, interface: type) -> Tuple[Binding, 'Binder']:
         is_scope = isinstance(interface, type) and issubclass(interface, Scope)
+        is_assisted_builder = _is_specialization(interface, AssistedBuilder)
         try:
-            return self._get_binding(interface, only_this_binder=is_scope)
+            return self._get_binding(interface, only_this_binder=is_scope or is_assisted_builder)
         except (KeyError, UnsatisfiedRequirement):
             if is_scope:
                 scope = interface
@@ -683,7 +665,7 @@ def _is_specialization(cls: type, generic_class: Any) -> bool:
     # We need to special-case Annotated as its __origin__ behaves differently than
     # other typing generic classes. See https://github.com/python/typing/pull/635
     # for some details.
-    if HAVE_ANNOTATED and generic_class is Annotated and isinstance(cls, _AnnotatedAlias):
+    if generic_class is Annotated and isinstance(cls, _AnnotatedAlias):
         return True
 
     if not hasattr(cls, '__origin__'):
@@ -1166,10 +1148,37 @@ class _BindingNotYetAvailable(Exception):
     pass
 
 
+# See a comment in _infer_injected_bindings() for why this is useful.
+class _NoReturnAnnotationProxy:
+    def __init__(self, callable: Callable) -> None:
+        self.callable = callable
+
+    def __getattribute__(self, name: str) -> Any:
+        # get_type_hints() uses quite complex logic to determine the namespaces using which
+        # any forward references should be resolved. Instead of mirroring this logic here
+        # let's just take the easy way out and forward all attribute access to the original
+        # callable except for the annotations – we want to filter them.
+        callable = object.__getattribute__(self, 'callable')
+        if name == '__annotations__':
+            annotations = callable.__annotations__
+            return {name: value for (name, value) in annotations.items() if name != 'return'}
+        return getattr(callable, name)
+
+
 def _infer_injected_bindings(callable: Callable, only_explicit_bindings: bool) -> Dict[str, type]:
+    def _is_new_union_type(instance: Any) -> bool:
+        new_union_type = getattr(types, 'UnionType', None)
+        return new_union_type is not None and isinstance(instance, new_union_type)
+
     spec = inspect.getfullargspec(callable)
+
     try:
-        bindings = get_type_hints(callable, include_extras=True)
+        # Return types don't matter for the purpose of dependency injection so instead of
+        # obtaining type hints of the callable directly let's wrap it in _NoReturnAnnotationProxy.
+        # The proxy removes the return type annotation (if present) from the annotations so that
+        # get_type_hints() works even if the return type is a forward reference that can't be
+        # resolved.
+        bindings = get_type_hints(cast(Callable, _NoReturnAnnotationProxy(callable)), include_extras=True)
     except NameError as e:
         raise _BindingNotYetAvailable(e)
 
@@ -1199,7 +1208,7 @@ def _infer_injected_bindings(callable: Callable, only_explicit_bindings: bool) -
 
         if only_explicit_bindings and _inject_marker not in metadata or _noinject_marker in metadata:
             del bindings[k]
-        elif _is_specialization(v, Union):
+        elif _is_specialization(v, Union) or _is_new_union_type(v):
             # We don't treat Optional parameters in any special way at the moment.
             union_members = v.__args__
             new_members = tuple(set(union_members) - {type(None)})
@@ -1209,7 +1218,20 @@ def _infer_injected_bindings(callable: Callable, only_explicit_bindings: bool) -
             # mypy complains about this construct:
             #     error: The type alias is invalid in runtime context
             # See: https://github.com/python/mypy/issues/5354
-            bindings[k] = new_union  # type: ignore
+            union_metadata = {
+                metadata
+                for member in new_members
+                for metadata in getattr(member, '__metadata__', tuple())
+                if _is_specialization(member, Annotated)
+            }
+            if (
+                only_explicit_bindings
+                and _inject_marker not in union_metadata
+                or _noinject_marker in union_metadata
+            ):
+                del bindings[k]
+            else:
+                bindings[k] = new_union  # type: ignore
 
     return bindings
 
