@@ -11,7 +11,7 @@
 """Functional tests for the "Injector" dependency injection framework."""
 
 from contextlib import contextmanager
-from typing import Any, NewType
+from typing import Any, NewType, Optional, Union
 import abc
 import sys
 import threading
@@ -25,7 +25,9 @@ import pytest
 from injector import (
     Binder,
     CallError,
+    Inject,
     Injector,
+    NoInject,
     Scope,
     InstanceProvider,
     ClassProvider,
@@ -46,12 +48,8 @@ from injector import (
     ClassAssistedBuilder,
     Error,
     UnknownArgument,
-    HAVE_ANNOTATED,
 )
 
-if HAVE_ANNOTATED:
-    from injector import Inject, NoInject
-
 
 class EmptyClass:
     pass
@@ -807,6 +805,19 @@ def test_assisted_builder_injection_is_safe_to_use_with_multiple_injectors():
     assert (b1._injector, b2._injector) == (i1, i2)
 
 
+def test_assisted_builder_injection_is_safe_to_use_with_child_injectors():
+    class X:
+        @inject
+        def __init__(self, builder: AssistedBuilder[NeedsAssistance]):
+            self.builder = builder
+
+    i1 = Injector()
+    i2 = i1.create_child_injector()
+    b1 = i1.get(X).builder
+    b2 = i2.get(X).builder
+    assert (b1._injector, b2._injector) == (i1, i2)
+
+
 class TestThreadSafety:
     def setup(self):
         self.event = threading.Event()
@@ -1449,38 +1460,75 @@ def test_get_bindings():
 
     assert get_bindings(function3b) == {'a': int}
 
-    if HAVE_ANNOTATED:
-        # The simple case of no @inject but injection requested with Inject[...]
-        def function4(a: Inject[int], b: str) -> None:
-            pass
+    # The simple case of no @inject but injection requested with Inject[...]
+    def function4(a: Inject[int], b: str) -> None:
+        pass
 
-        assert get_bindings(function4) == {'a': int}
+    assert get_bindings(function4) == {'a': int}
 
-        # Using @inject with Inject is redundant but it should not break anything
-        @inject
-        def function5(a: Inject[int], b: str) -> None:
-            pass
+    # Using @inject with Inject is redundant but it should not break anything
+    @inject
+    def function5(a: Inject[int], b: str) -> None:
+        pass
 
-        assert get_bindings(function5) == {'a': int, 'b': str}
+    assert get_bindings(function5) == {'a': int, 'b': str}
 
-        # We need to be able to exclude a parameter from injection with NoInject
-        @inject
-        def function6(a: int, b: NoInject[str]) -> None:
-            pass
+    # We need to be able to exclude a parameter from injection with NoInject
+    @inject
+    def function6(a: int, b: NoInject[str]) -> None:
+        pass
 
-        assert get_bindings(function6) == {'a': int}
+    assert get_bindings(function6) == {'a': int}
 
-        # The presence of NoInject should not trigger anything on its own
-        def function7(a: int, b: NoInject[str]) -> None:
-            pass
+    # The presence of NoInject should not trigger anything on its own
+    def function7(a: int, b: NoInject[str]) -> None:
+        pass
 
-        assert get_bindings(function7) == {}
+    assert get_bindings(function7) == {}
 
-        # There was a bug where in case of multiple NoInject-decorated parameters only the first one was
-        # actually made noninjectable and we tried to inject something we couldn't possibly provide
-        # into the second one.
-        @inject
-        def function8(a: NoInject[int], b: NoInject[int]) -> None:
-            pass
+    # There was a bug where in case of multiple NoInject-decorated parameters only the first one was
+    # actually made noninjectable and we tried to inject something we couldn't possibly provide
+    # into the second one.
+    @inject
+    def function8(a: NoInject[int], b: NoInject[int]) -> None:
+        pass
 
-        assert get_bindings(function8) == {}
+    assert get_bindings(function8) == {}
+
+    # Default arguments to NoInject annotations should behave the same as noninjectable decorator w.r.t 'None'
+    @inject
+    @noninjectable('b')
+    def function9(self, a: int, b: Optional[str] = None):
+        pass
+
+    @inject
+    def function10(self, a: int, b: NoInject[Optional[str]] = None):
+        # b:s type is Union[NoInject[Union[str, None]], None]
+        pass
+
+    assert get_bindings(function9) == {'a': int} == get_bindings(function10)
+
+    # If there's a return type annottion that contains an a forward reference that can't be
+    # resolved (for whatever reason) we don't want that to break things for us – return types
+    # don't matter for the purpose of dependency injection.
+    @inject
+    def function11(a: int) -> 'InvalidForwardReference':
+        pass
+
+    assert get_bindings(function11) == {'a': int}
+
+
+# Tests https://github.com/alecthomas/injector/issues/202
+@pytest.mark.skipif(sys.version_info < (3, 10), reason="Requires Python 3.10+")
+def test_get_bindings_for_pep_604():
+    @inject
+    def function1(a: int | None) -> None:
+        pass
+
+    assert get_bindings(function1) == {'a': int}
+
+    @inject
+    def function1(a: int | str) -> None:
+        pass
+
+    assert get_bindings(function1) == {'a': Union[int, str]}
